#!/usr/bin/env python3

# ----------------------------------------------------------------------------
# Copyright (c) 2020--, Qiyun Zhu.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

__name__ = 'woltka'
__description__ = 'versatile metagenome classifier'
__version__ = '0.1.4'
__license__ = 'BSD-3-Clause'
__author__ = 'Qiyun Zhu'
__email__ = 'qiyunzhu@gmail.com'
__url__ = 'https://github.com/qiyunzhu/woltka'
